/*====================================================================
 Version: June 03, 2008
\===================================================================*/

/*====================================================================
 Matthieu Guerquin-Kern

 EPFL/STI/IMT/LIB
 BM 4.140 (Bâtiment BM)
 Station 17
 CH-1015 Lausanne VD
 Switzerland

 phone:	+41 (0) 21 6935142
 E-mail:	matthieu.guerquin-kern@epfl.ch
 URL:		http://bigwww.epfl.ch/guerquin
\===================================================================*/

/*====================================================================
 This work is based on a personnal unpublished draft.

 However, one can refer to:

 F. Wajer, K. P. Pruessmann,
 Major speedup of reconstruction for sensitivity encoding with arbitrary trajectories,
 in Proc., ISMRM, 8th Annual Meeting, Glasgow, 2001, p. 767.

\===================================================================*/

/*====================================================================
 Copyright free:

 You'll be free to use this software for research purposes, but you
 should not redistribute it without our consent. In addition, we expect
 you to include a citation or acknowledgement whenever you present or
 publish results that are based on it.
\===================================================================*/

import ij.*;
import ij.gui.*;
import ij.plugin.filter.ExtendedPlugInFilter;
import ij.plugin.filter.PlugInFilterRunner;
import ij.process.*;
import ij.util.*;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Panel;
import java.awt.TextArea;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;

/*====================================================================
	MriConvolutionKernels_
\===================================================================*/

/*********************************************************************
 * This plugin implements the operation of kspace sampling (as in MRI),
 * followed by its adjoint. You can choose the kspace trajectory type
 * and you'll be returned the image and the convolution kernel.
 * <br><br>
 * <i>You'll be free to use this software for research purposes, but you
 * should not redistribute it without our consent.  In addition, we expect
 * you to include a citation or acknowledgement whenever you present or
 * publish results that are based on it.</i>
 * @author Matthieu Guerquin-Kern
 * @version June 03, 2008
 * @see MRITools_matt
 *********************************************************************/
public class MriConvolutionKernels_ implements ExtendedPlugInFilter {

	private static final int CAPABILITIES = DOES_8G | DOES_16 | DOES_32 | CONVERT_TO_FLOAT;

	/*********************************************************************
	 * Dialog items for the k-space trajectory.
	 ********************************************************************/
	protected static final String[] TRAJECTORIES = {
		"Full_Cartesian",
		"Undersampled_Cartesian",
		"Full_Spiral",
		"Undersampled_Spiral",
		"Full_Random",
		"Undersampled_Random"
	};

	/*********************************************************************
 Dialog label.
 @see #TRAJECTORIES
	 ********************************************************************/
	protected static final String KSPACE_TRAJECTORY = "K-space trajectory";

	private final GenericDialog dialog = new GenericDialog("MriConvKernels");

	ImagePlus imp;

	/*....................................................................
	ExtendedPlugInFilter methods
....................................................................*/

	/*********************************************************************
 This method is the first one to be called by ImageJ. It is used to
 query this plugin about its capabilities. It is also the last method
 to be called, with <code>arg</code> set to
 <code>&#34;final&#34;</code>.
 @param arg A <code>String</code> that contains arguments to this
 plugin.
 @param imp The <code>ImagePlus</code> image currently selected, or
 <code>null</code> if there is none.
	 ********************************************************************/
	public int setup ( final String arg, final ImagePlus imp ) {
		this.imp = imp;
		return(CAPABILITIES);
	} /* end setup */

	/*********************************************************************
 This method is called by ImageJ to inform this plugin of the number
 of passes. Unused in this plugin.
 @param nPasses The number of passes.
	 ********************************************************************/
	public void setNPasses ( final int nPasses ) { } /* end setNPasses */

	/*********************************************************************
 This method is called by ImageJ to process the image.
 @param ip The <code>ImageProcessor</code> to process. It is expected
 to extend <code>FloatProcessor</code>, but this convention is not
 tested at run-time.
	 ********************************************************************/
	public void run ( final ImageProcessor ip ) {
		final int width = ip.getWidth();
		final int height = ip.getHeight();
		
		// Check if the dimensions are power fo two 
		int w=2;
		while(w<width) w *= 2;
		int h=2;
		while(h<height) h *= 2;
		if ((w!=width)||(h!=height)) {
			IJ.error("The input image should have dimensions that are power of 2");
			return;
		}
		
		// Allocate memory
		IJ.showProgress(0);
		IJ.showStatus("Memory allocation...");
		double[] dataReal = Tools.toDouble((float[])ip.getPixels());
		double[] dataImag = new double[width*height];

		final String KspaceTrajAsString = dialog.getNextChoice();

		IJ.showProgress(0.1);
		IJ.showStatus("Generating the trajectory...");
		MRITools_matt exp = new MRITools_matt(width,height,KspaceTrajAsString);

		exp.showTraj();
		IJ.showProgress(0.2);

		IJ.showStatus("Building the corresponding kernel...");
		exp.ComputeMriKernel(2);
		exp.PlotKernel();
		IJ.showProgress(0.6);

		IJ.showStatus("Performing the convolution...");
		exp.MRIConvolution(dataReal,dataImag);
		IJ.showProgress(0.9);

		ImageProcessor ip3 = new FloatProcessor( width, height);
		ip3.setPixels(Tools.toFloat(dataReal));
		final ImagePlus imp3 = new ImagePlus("Real part of the Fourier transform",ip3);
		imp3.resetDisplayRange();
		imp3.show();

		IJ.showProgress(1);
	} /* end run */

	/*********************************************************************
 This method is called by ImageJ to let this plugin interact with the
 user.
 @param imp The <code>ImagePlus</code> image currently selected, or
 <code>null</code> if there is none, but an absence of image is not
 tested at run-time.
 @param command A <code>String</code> that contains arguments to this
 plugin.
 @param pfr The parent <code>PlugInFilterRunner</code> that called
 this method.
	 ********************************************************************/
	public int showDialog ( final ImagePlus imp, final String command, final PlugInFilterRunner pfr ) {
		dialog.addChoice(KSPACE_TRAJECTORY, TRAJECTORIES, TRAJECTORIES[3]);
		dialog.addPanel(new MriConvolutionKernelsCreditsButton());
		dialog.showDialog();
		if (dialog.wasCanceled()) {
			return(DONE);
		}
		if (dialog.wasOKed()) {
			return(CAPABILITIES);
		}
		else {
			return(DONE);
		}
	} /* end showDialog */

} // End MriConvolutionKernels_

/*====================================================================
	MriConvolutionKernelsCredits
\===================================================================*/

/*********************************************************************
 The purpose of this class is to manage the panel that displays the
 list of credits for this citationware plugin. For convenience, it
 takes advantage of <code>java.swing</code> classes. Note that, at the
 time of writing, this is not in line with the conventions of ImageJ,
 but it is expected that <code>java.swing</code> will ultimately
 supersede <code>java.awt</code> in the long run, even within ImageJ.
 @see MriConvolutionKernelsCreditsButton
 ********************************************************************/
class MriConvolutionKernelsCredits	extends JDialog implements ActionListener{
	/* begin class MriConvolutionKernelsCredits */

	/*....................................................................
	MriConvolutionKernelsCredits constructors
....................................................................*/

	/**
	 * 
	 */
	private static final long serialVersionUID = -8240232076031910029L;

	/*********************************************************************
 This constructor prepares the dialog and displays it.
 @param parentWindow The <code>JFrame</code> object that is the parent
 of this dialog.
	 ********************************************************************/
	protected MriConvolutionKernelsCredits ( final JFrame parentWindow) {
		super(parentWindow, "MriConvolutionKernels_", true);
		getContentPane().setLayout(new BorderLayout(0, 20));
		final JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
		final JButton doneButton = new JButton("Done");
		doneButton.addActionListener(this);
		buttonPanel.add(doneButton);
		final TextArea text = new TextArea(30, 50);
		text.setEditable(false);
		text.append("\nVersion: June 03, 2008\n\n");
		text.append("Matthieu Guerquin-Kern\n\n");
		text.append("EPFL/STI/IMT/LIB\n");
		text.append("BM 4.140 (Batiment BM)\n");
		text.append("Station 17\n");
		text.append("CH-1015 Lausanne VD\n");
		text.append("Switzerland\n\n");
		text.append("phone:		+41 (0) 21 6935142\n");
		text.append("E-mail:		matthieu.guerquin-kern@epfl.ch\n");
		text.append("URL:		http://bigwww.epfl.ch/guerquin\n\n");
		text.append("This work is based on a personnal unpublished draft.\n");
		text.append("However, one can refer to:\n\n");
		text.append("	F. Wajer, K. P. Pruessmann,\n");
		text.append("	'Major speedup of reconstruction for sensitivity\n");
		text.append("	encoding with arbitrary trajectories,'\n");
		text.append("	in Proc., ISMRM, 8th Annual Meeting, Glasgow, 2001, p. 767.\n\n");
		text.append("Copyright free:\n\n");
		text.append("You'll be free to use this software for research purposes, but you\n");
		text.append("should not redistribute it without our consent. In addition, we expect\n");
		text.append("you to include a citation or acknowledgement whenever you present or\n");
		text.append("publish results that are based on it.");
		getContentPane().add("Center", text);
		getContentPane().add("South", buttonPanel);
		pack();
		GUI.center(this);
		setVisible(true);
	} /* end MriConvolutionKernelsCredits */

	/*....................................................................
	ActionListener methods
....................................................................*/

	/*********************************************************************
 This method implements the <code>ActionListener</code> that deals
 with the only button of the dialog. Used to close the credits panel.
 @param e The <code>ActionEvent</code> event. Only one event is
 of interest here.
	 ********************************************************************/
	public void actionPerformed (final ActionEvent e) {
		if (e.getActionCommand().equals("Done")) {
			dispose();
		}
	} /* end actionPerformed */

} /* end class MriConvolutionKernelsCredits */

/*====================================================================
	MriConvolutionKernelsCreditsButton
\===================================================================*/

/*********************************************************************
 The purpose of this class is to encapsulate a button inside of a
 <code>Panel</code> object, so that it can be added to the
 <code>GenericDialog</code> object created in the method
 <code>MriConvolutionKernels_.showDialog()</code>. The button triggers the
 display of the credits. For convenience, this class takes advantage
 of <code>java.swing</code> classes. Note that, at the time of
 writing, this is not in line with the conventions of ImageJ, but it
 is expected that <code>java.swing</code> will ultimately supersede
 <code>java.awt</code> in the long run, even within ImageJ.
 @see MriConvolutionKernels_#showDialog(ImagePlus, String, PlugInFilterRunner)
 @see MriConvolutionKernelsCreditsButton
 ********************************************************************/
class MriConvolutionKernelsCreditsButton extends	Panel implements ActionListener {
	/* begin class MriConvolutionKernelsCreditsButton */

	/*....................................................................
	MriConvolutionKernelsCreditsButton constructors
....................................................................*/

	/**
	 * 
	 */
	private static final long serialVersionUID = -8030153272800727381L;

	/*********************************************************************
 This constructor prepares the panel with one button and its listener.
	 ********************************************************************/
	protected MriConvolutionKernelsCreditsButton () {
		super();
		final JButton creditsButton = new JButton("Credits");
		creditsButton.addActionListener(this);
		add(creditsButton);
	} /* end MriConvolutionKernelsCreditsButton */

	/*....................................................................
	ActionListener methods
....................................................................*/

	/*********************************************************************
 This method triggers the display of the credits.
 @param e The <code>ActionEvent</code> event. Only one event is
 expected here.
 @see MriConvolutionKernelsCredits
	 ********************************************************************/
	public void actionPerformed ( final ActionEvent e ) {
		new MriConvolutionKernelsCredits(null);
	} /* end actionPerformed */

} /* end class MriConvolutionKernelsCreditsButton */


/*====================================================================
	MRITools_matt
\===================================================================*/

/*********************************************************************
 * This class offers tools for the ImageJ plugin MriConvolutionKernels_ .
 * 
 * <br><br>
 * <i>You'll be free to use this software for research purposes, but you
 * should not redistribute it without our consent.  In addition, we expect
 * you to include a citation or acknowledgement whenever you present or
 * publish results that are based on it.</i>
 * @author Matthieu Guerquin-Kern
 * @version June 03, 2008
 * @see MriConvolutionKernels_
 * @see KSpaceTrajectory
 * @see FFTTools_matt
 *********************************************************************/
class MRITools_matt  extends MriConvolutionKernels_{

	private int width;
	private int height;
	private KSpaceTrajectory k;
	private double[] G_real;
	private double[] G_imag;

	/**
	 * First constructor.
	 * @param w width
	 * @param h height
	 */
	MRITools_matt(int w, int h){
		width=2;
		while(width<w) width *= 2;
		height=2;
		while(height<w) height *= 2;
	}

	/**
	 * Second constructor.
	 * @param w width
	 * @param h height
	 * @param type type of the K-space trajectory
	 */
	MRITools_matt(int w, int h, String type){
		width=2;
		while(width<w) width *= 2;
		height=2;
		while(height<h) height *= 2;

		if (type.equals(TRAJECTORIES[0])) {
			k = new FullCartesianTraj(w,h);
		}else if (type.equals(TRAJECTORIES[1])) {
			k = new UndersampledCartesianTraj(w,h);
		}else if (type.equals(TRAJECTORIES[2])) {
			k = new FullSpiralTraj(w,h);
		} else if (type.equals(TRAJECTORIES[3])) {
			k = new UndersampledSpiralTraj(w,h);
		} else if (type.equals(TRAJECTORIES[4])) {
			k = new FullRandomTraj(w,h);
		} else if (type.equals(TRAJECTORIES[5])) {
			k = new UndersampledRandomTraj(w,h);
		} else{
			System.out.println("unknown trajectory");
		}
	}

	/**
	 * @return width
	 */
	public int getWidth(){
		return width;
	}

	/**
	 * @return height
	 */
	public int getHeight(){
		return height;
	}

	/**
	 * Calls internal KSpaceTrajectory function that displays trajectory.
	 */
	public void showTraj(){
		k.PlotKSpaceTraj();
	}

	/**
	 * Computes the kernel associated with the trajectory. Stores the kernel in the class fields.
	 * @param refinement factor of refinement for the FT Cartesian grid
	 * @see FFTTools_matt 
	 */
	public void ComputeMriKernel(int refinement) {
		G_real = new double[(2*width)*(2*height)];
		G_imag = new double[(2*width)*(2*height)];

		double[] kspace1_real = new double[width*height*refinement*refinement];
		double[] kspace1_imag = new double[width*height*refinement*refinement];
		double[] kspace2_real = new double[width*height*refinement*refinement];
		double[] kspace2_imag = new double[width*height*refinement*refinement];
		int indx1,indy1;
		int indx2,indy2;
		int index1,index2;

		for (int i=0;i < k.Nb_samples;i++){
			indx1 = (int)modulo((int)Math.round((double)refinement*k.kx[i]) , (double)(refinement*width));
			indy1 = (int)modulo((int)Math.round((double)refinement*k.ky[i]) , (double)(refinement*height));
			indx2 = (int)modulo((int)Math.round(-(double)refinement*k.kx[i]) , (double)(refinement*width));
			indy2 = indy1;
			index1 = indx1 + refinement*width*indy1;
			index2 = indx2 + refinement*width*indy2;
			kspace1_real[index1] += 1.0/((double)(width*height));///Math.pow(refinement,2);
			kspace2_real[index2] += 1.0/((double)(width*height));///Math.pow(refinement,2);
		}

		FFTTools_matt.doFFT( kspace1_real , kspace1_imag , width*refinement , height*refinement);
		FFTTools_matt.doFFT( kspace2_real , kspace2_imag , width*refinement , height*refinement);

		for (int i=0;i<width;i++){
			for (int j=0;j<height;j++){
				index1 = i  + (2*width)*(j);
				index2 = i + width*refinement*j;
				G_real[index1] =  kspace1_real[index2];
				G_imag[index1] =  kspace1_imag[index2];

				if ((i==0)||(j==0)){
					if ((i==0)&&(j!=0)){
						index1 = 0 + (2*width)*(j+height);
						index2 = 0 + width*refinement*(height-j);
						G_real[index1] = kspace2_real[index2];
						G_imag[index1] = kspace2_imag[index2];
					} else{if (i!=0){
						index1 = i + width + (2*width)*(0);
						index2 = (width-i) + width*refinement*(0);
						G_real[index1] =  kspace1_real[index2];
						G_imag[index1] = -kspace1_imag[index2];
					}}
				} else{
					index1 = i + (2*width)*(j+height);
					index2 = i + width*refinement*(height-j);
					G_real[index1] = kspace2_real[index2];
					G_imag[index1] = kspace2_imag[index2];

					index1 = i + width + (2*width)*(j+height);
					index2 = (width-i) + width*refinement*(height-j);
					G_real[index1] =  kspace1_real[index2];
					G_imag[index1] = -kspace1_imag[index2];

					index1 = i + width + (2*width)*(j);
					index2 = (width-i) + width*refinement*(j);
					G_real[index1] =  kspace2_real[index2];
					G_imag[index1] = -kspace2_imag[index2];
				}
			}
		}
	}
	/**
	 * Plots as an image the kernel stored in class fields.
	 */
	public void PlotKernel() {
		float[] data = new float[(2*width-1)*(2*height-1)];
		int index1,index2;

		// Perform an quadrant swap and computes the modulus
		for (int i=0;i<width;i++){
			for (int j=0;j<height;j++) {
				index2 = i  + (2*width)*(j);
				index1 = i+width-1  + (2*width-1)*(j+height-1);
				data[index1] = modulus(G_real[index2],G_imag[index2]);
				if ((i==0)||(j==0)){
					if (j!=0){ // and i==0
						index2 = 0 + (2*width)*(j+height);
						index1 = 0 + width-1 + (2*width-1)*(j-1);
						data[index1] = modulus(G_real[index2],G_imag[index2]);
					}else if (i!=0){  // and j==0
						index2 = i+width  + (2*width)*(0);
						index1 = i-1 + (2*width-1)*(0 +width-1);
						data[index1] = modulus(G_real[index2],G_imag[index2]);
					}
				} else{
					index2 = i+width  + (2*width)*(j+height);
					index1 = i-1 + (2*width-1)*(j-1);
					data[index1] = modulus(G_real[index2],G_imag[index2]);

					index2 = i  + (2*width)*(j+height);
					index1 = i+width-1  + (2*width-1)*(j-1);
					data[index1] = modulus(G_real[index2],G_imag[index2]);

					index2 = i+width  + (2*width)*(j);
					index1 = i-1  + (2*width-1)*(j+height-1);
					data[index1] = modulus(G_real[index2],G_imag[index2]);
				}
			}
		}

		ImageProcessor ip2 = new FloatProcessor( 2*width-1, 2*height-1 );
		ip2.setPixels(data);
		ip2.resetMinAndMax();
		final ImagePlus imp2 = new ImagePlus("Convolution Kernel (Modulus)",ip2);
		imp2.resetDisplayRange();
		imp2.show();
	}

	/**
	 * Performs the convolution of an image with the kernel.
	 * @param Image_real real part of the image
	 * @param Image_imag imaginary part of the image
	 * @see FFTTools_matt
	 */
	public void MRIConvolution( double[] Image_real, double[] Image_imag ) {
		FFTTools_matt.doFFT( G_real , G_imag , 2*width , 2*height);
		//PlotKernel();
		double[] real = new double[(2*width)*(2*height)];
		double[] imag = new double[(2*width)*(2*height)];
		int index1,index2;

		for (int i=0;i<width;i++){
			for (int j=0;j<height;j++) {
				index1 = i+(2*width)*j;
				index2 = i+width*j;
				real[index1] = Image_real[index2];
				imag[index1] = Image_imag[index2];
			}
		}
		FFTTools_matt.doFFT( real , imag , 2*width , 2*height);

		for (int i=0;i<2*width;i++){
			for (int j=0;j<2*height;j++) {
				index1 = i+(2*width)*j;
				real[index1] = G_real[index1]*real[index1]-G_imag[index1]*imag[index1];
				imag[index1] = G_real[index1]*imag[index1]+G_imag[index1]*real[index1];
			}
		}
		FFTTools_matt.inverseFFT( real , imag , 2*width , 2*height);

		for (int i=0;i<width;i++){
			for (int j=0;j<height;j++) {
				index1 = i+(2*width)*j;
				index2 = i+width*j;
				Image_real[index2] = real[index1];
				Image_imag[index2] = imag[index1];
			}
		}

		//FFTTools_matt.inverseFFT( G_real , G_imag , 2*width , 2*height);
	}

	private double modulo(double x, double a){
		return x-a*Math.floor(x/a);
	}

	private float modulus(double real, double imag){
		return (float)Math.sqrt(Math.pow(real,2)+Math.pow(imag,2));
	}

} // End MRITools_matt

/*====================================================================
	KSpaceTrajectory
\===================================================================*/

/**
 * Super Class that describes an MRI k-space trajectory.
 * @see MRITools_matt
 */
class KSpaceTrajectory{
	protected int width;
	protected int height;
	public double[] kx;
	public double[] ky;
	public int Nb_samples;

	KSpaceTrajectory (int w, int h){
		width = w;
		height = h;
	}

	public String type() {
		return "KSpaceTrajectory";
	}

	public double getKx(int i){
		return kx[i];
	}

	public double getKy(int i){
		return ky[i];
	}

	public int getNbSamples(){
		return Nb_samples;
	}

	public void PlotKSpaceTraj() {
		ImageProcessor ip2 = new ByteProcessor( 2*width+1, 2*height+1 );
		final byte[] kspace = new byte[(2*width+1)*(2*height+1)];
		int indx,indy,index;

		for (int i=0;i<Nb_samples;i++){
			indx = (int)Math.floor(2*(kx[i]+(double)width/2))+1;
			indy = (int)Math.floor(2*(ky[i]+(double)height/2))+1;
			index = indx + (2*width+1)*indy;
			kspace[index] = (byte)255;
		}
		ip2.setPixels(kspace);
		ip2.invert();
		ip2.drawLine(width,0,width,2*height+1);
		ip2.drawLine(0,height,2*width+1,height);
		ip2.drawString("kx",2*width-20,height-3);
		ip2.drawString("ky",width-20,15);
		ip2.drawString("O",width-12,height+15);
		ip2.drawString("(c) BIG - 2008",2*width-100,2*height-2);
		final ImagePlus imp2 = new ImagePlus("K-space trajectory",ip2);
		imp2.show();
		imp2.updateAndDraw();
	}

} // KSpaceTrajectory

/*====================================================================
	FullCartesianTraj
\===================================================================*/
class FullCartesianTraj extends KSpaceTrajectory {
	FullCartesianTraj(int w, int h) {
		super(w,h);
		Nb_samples = width*height;
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];
		int ind_sample = 0;
		int kyinf = -(int)Math.ceil((double)height/2);
		int kysup = (int)Math.floor((double)height/2);
		int kxinf = -(int)Math.ceil((double)width/2);
		int kxsup = (int)Math.floor((double)width/2);

		for (int j=kysup;j>kyinf;j--){
			for (int i=kxinf;i<kxsup;i++){
				kx[ind_sample] = (double)i;
				ky[ind_sample] = -(double)j;
				ind_sample++;
			}
		}
	}

	public String type() {
		return "FullCartesianTraj";
	}
} // FullCartesianTraj

/*====================================================================
	UndersampledCartesianTraj
\===================================================================*/
class UndersampledCartesianTraj extends KSpaceTrajectory {
	UndersampledCartesianTraj(int w, int h) {
		super(w,h);
		int kyinf = -(int)Math.ceil((double)height/2);
		int kysup = 2*(int)Math.floor((double)height/4);
		int kxinf = -(int)Math.ceil((double)width/2);
		int kxsup = (int)Math.floor((double)width/2);

		Nb_samples = 0;
		for (int j=kysup;j>=0;j-=2){
			for (int i=kxinf;i<kxsup;i++){
				Nb_samples++;
			}
		}
		for (int j=-1;j>kyinf;j-=2){
			for (int i=kxinf;i<kxsup;i++){
				Nb_samples++;
			}
		}
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];
		int ind_sample = 0;

		for (int j=kysup;j>=0;j-=2){
			for (int i=kxinf;i<kxsup;i++){
				kx[ind_sample] =  (double)i;
				ky[ind_sample] = -(double)j;
				ind_sample++;
			}
		}
		for (int j=-1;j>kyinf;j-=2){
			for (int i=kxinf;i<kxsup;i++){
				kx[ind_sample] =  (double)i;
				ky[ind_sample] = -(double)j;
				ind_sample++;
			}
		}
	}

	public String type() {
		return "UndersampledCartesianTraj";
	}
} // UndersampledCartesianTraj

/*====================================================================
	FullSpiralTraj
\===================================================================*/
class FullSpiralTraj extends KSpaceTrajectory {
	FullSpiralTraj(int w, int h) {
		super(w,h);
		int res;

		if (width!=height){
			res = width;
		}else{
			res = Math.min(width,height);
			System.out.println("Warning: Spiral trajectories are better suited for square images!");
		}
		double kmax = Math.floor((double)res/2)-1;
		double Dk = 1;
		double A = Dk/2/Math.PI;
		double Nr = kmax/Dk;
		double omega = 2*Math.PI/Math.sqrt(1+Math.pow(2*Math.PI*kmax*10/Dk,2));
		int Tend = (int)(2*Math.PI*Nr/omega);
		double[] phi = new double[Tend];
		double[] k_tmp_real = new double[Tend];
		double[] k_tmp_imag = new double[Tend];

		// we first generate a spiral that has a non uniform density
		for (int t=0;t<Tend;t++){
			phi[t] = (double)(omega*t);
			k_tmp_real[t] = A * phi[t] * Math.cos(phi[t]);
			k_tmp_imag[t] = A* phi[t] * Math.sin(phi[t]);
		}        

		// we try to sample the first traj so that we have a near Nyquist density
		// This first loop determines the number of samples in the final traj
		int counter = 0;
		double k_prec_real = k_tmp_real[0];
		double k_prec_imag = k_tmp_imag[0];
		double distance;
		for (int ind_k=0;ind_k<Tend;ind_k++){
			distance = Math.sqrt(Math.pow(k_tmp_real[ind_k]-k_prec_real,2)+Math.pow(k_tmp_imag[ind_k]-k_prec_imag,2));
			if (((distance>0.5*Dk)&&(phi[ind_k]<2*Math.PI))||((distance>Dk)&&(phi[ind_k]>2*Math.PI))){
				counter++;
				k_prec_real = k_tmp_real[ind_k];
				k_prec_imag = k_tmp_imag[ind_k];
			}
		}
		Nb_samples = counter+1;
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];

		// This second loop fills the final traj
		counter = 0;
		kx[0] = k_tmp_real[0]*width/res;
		ky[0] = k_tmp_imag[0]*height/res;
		for (int ind_k=0;ind_k<Tend;ind_k++){
			distance = Math.sqrt(Math.pow(k_tmp_real[ind_k]-kx[counter]*res/width,2)+Math.pow(k_tmp_imag[ind_k]-ky[counter]*res/height,2));
			if (((distance>0.5*Dk)&&(phi[ind_k]<2*Math.PI))||((distance>Dk)&&(phi[ind_k]>2*Math.PI))){
				counter++;
				kx[counter] = k_tmp_real[ind_k]*width/res;
				ky[counter] = k_tmp_imag[ind_k]*height/res;
			}
		}
	}

	public String type() {
		return "FullSpiralTraj";
	}
} // FullSpiralTraj

/*====================================================================
	UndersampledSpiralTraj
\===================================================================*/
class UndersampledSpiralTraj extends KSpaceTrajectory {
	UndersampledSpiralTraj(int w, int h) {
		super(w,h);
		int res;

		if (width!=height){
			res = width;
		}else{
			res = Math.min(width,height);
			System.out.println("Warning: Spiral trajectories are better suited for square images!");
		}

		double kmax = Math.floor((double)res/2)*0.99;
		double Dk = 1;
		double kmax1 =  Math.floor(kmax/3);
		double A1 = Dk/2/Math.PI;
		double phi0 = (kmax1/2-Math.floor(kmax1/2))/A1;
		double Nr1 = kmax1/Dk;
		double omega1 = 2*Math.PI/Math.sqrt(1+Math.pow(2*Math.PI*kmax1*10/Dk,2));
		int Tend1 = (int)(2*Math.PI*Nr1/omega1);

		double A2 = Dk/Math.PI;
		double omega2 = 2*Math.PI/Math.sqrt(1+Math.pow(2*Math.PI*kmax*10/Dk,2));
		int Tend = Tend1+(int)((kmax-kmax1)*Math.PI/omega2);
		double[] phi = new double[Tend];

		double[] k_tmp_real = new double[Tend];
		double[] k_tmp_imag = new double[Tend];

		// we first generate a spiral that has a non uniform density
		for (int t=0;t<Tend1;t++){
			phi[t] = (double)(omega1*t);
			k_tmp_real[t] = A1 * phi[t] * Math.cos(phi[t]-phi0-Math.PI/2);
			k_tmp_imag[t] = A1 * phi[t] * Math.sin(phi[t]-phi0-Math.PI/2);
		}
		for (int t=Tend1;t<Tend;t++){
			phi[t] = (double)(omega2*t);
			k_tmp_real[t] = (A2 * (phi[t]-omega2*Tend1)+kmax1) * Math.cos(phi[t]-omega2*Tend1-Math.PI/2);
			k_tmp_imag[t] = (A2 * (phi[t]-omega2*Tend1)+kmax1) * Math.sin(phi[t]-omega2*Tend1-Math.PI/2);
		}       

		// we try to sample the first traj so that we have a near Nyquist density
		// This first loop determines the number of samples in the final traj
		int counter = 0;
		double k_prec_real = k_tmp_real[0];
		double k_prec_imag = k_tmp_imag[0];
		double distance;
		for (int ind_k=0;ind_k<Tend;ind_k++){
			distance = Math.sqrt(Math.pow(k_tmp_real[ind_k]-k_prec_real,2)+Math.pow(k_tmp_imag[ind_k]-k_prec_imag,2));
			if (((distance>0.5*Dk)&&(phi[ind_k]<2*Math.PI))||((distance>Dk)&&(phi[ind_k]>2*Math.PI))){
				counter++;
				k_prec_real = k_tmp_real[ind_k];
				k_prec_imag = k_tmp_imag[ind_k];
			}
		}
		Nb_samples = counter+1;
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];

		// This second loop fills the final traj
		counter = 0;
		kx[0] = k_tmp_real[0]*width/res;
		ky[0] = k_tmp_imag[0]*height/res;
		for (int ind_k=0;ind_k<Tend;ind_k++){
			distance = Math.sqrt(Math.pow(k_tmp_real[ind_k]-kx[counter]*res/width,2)+Math.pow(k_tmp_imag[ind_k]-ky[counter]*res/height,2));
			if (((distance>0.5*Dk)&&(phi[ind_k]<2*Math.PI))||((distance>Dk)&&(phi[ind_k]>2*Math.PI))){
				counter++;
				kx[counter] = k_tmp_real[ind_k]*width/res;
				ky[counter] = k_tmp_imag[ind_k]*height/res;
			}
		}
	}

	public String type() {
		return "UndersampledSpiralTraj";
	}
} // UndersampledSpiralTraj

/*====================================================================
	FullRandomTraj
\===================================================================*/
class FullRandomTraj extends KSpaceTrajectory {
	FullRandomTraj(int w, int h) {
		super(w,h);
		Nb_samples = width*height;
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];
		for (int i=0;i<Nb_samples;i++){
			kx[i] = (Math.random() - 0.5) * width;
			ky[i] = (Math.random() - 0.5) * height;
		}
	}

	public String type() {
		return "FullRandomTraj";
	}
} // FullRandomTraj

/*====================================================================
	UndersampledRandomTraj
\===================================================================*/
class UndersampledRandomTraj extends KSpaceTrajectory {
	UndersampledRandomTraj(int w, int h) {
		super(w,h);
		Nb_samples = (int)Math.floor(0.3*(double)width*(double)height);
		kx = new double[Nb_samples];
		ky = new double[Nb_samples];
		for (int i=0;i<Nb_samples;i++){
			kx[i] = (Math.random() - 0.5) * width;
			ky[i] = (Math.random() - 0.5) * height;
		}
	}

	public String type() {
		return "UndersampledRandomTraj";
	}
} // UndersampledRandomTraj


/*====================================================================
This class offers tools to perform Fourier transform on images.

Adapted from:

 * IP-LAB
 * ImageProcessing Laboratories
 *
 * Freely available software and computer sessions for teaching and programming image processing in Java.
 *
 * Reference
 * D. Sage, M. Unser
 * "Teaching Image-Processing Programming in Java,"
 * IEEE Signal Processing Magazine, vol. 20, no. 6, pp. 43-52, November 2003.
 *
 * Author: Daniel Sage
 * Organisation: Biomedical Image Group (BIG), Swiss Federal Institute of Teachnology Lausanne (EPFL)

\===================================================================*/

/*====================================================================
	FFTTools_matt
\===================================================================*/

/**********************************************************************
This class offers tools to perform Fourier transform on images.
<br><br>
Adapted from:
<br><br>
 * IP-LAB<br>
 * ImageProcessing Laboratories<br>
 *<br>
 * Freely available software and computer sessions for teaching and programming image processing in Java.<br>
 *<br>
 * Reference<br>
 * D. Sage, M. Unser<br>
 * "Teaching Image-Processing Programming in Java,"<br>
 * IEEE Signal Processing Magazine, vol. 20, no. 6, pp. 43-52, November 2003.<br>
 *<br>
 * Author: Daniel Sage<br>
 * Organisation: Biomedical Image Group (BIG), Swiss Federal Institute of Teachnology Lausanne (EPFL)
 ***********************************************************************/
class FFTTools_matt {

	/*------------------------------------------------------------------*/
	/**
	 * Perform the FFT 2D with the method Decimation in Time
	 * Each image's dimension has to be a power of two : 2**m
	 *
	 * @param imageReal  input: Signal, output: Fourier coefficients (real part) 
	 * @param imageImag  input: Signal, output: Fourier coefficients (imaginary part)
	 * @param width		width of the signal
	 * @param height		height of the signal
	 */
	static public void doFFT( double[] imageReal, double[] imageImag, int width, int height) {
		double colReal[] = new double[height];
		double colImag[] = new double[height];
		for (int x=0;x<width;x++) {
			extractColumn( imageReal , width , x , colReal );	
			extractColumn( imageImag , width , x , colImag );						
			doFFT1D( colReal , colImag );
			putColumn( imageReal , width , x , colReal );	
			putColumn( imageImag , width , x , colImag );
		}

		double rowReal[] = new double[width];
		double rowImag[] = new double[width];
		for (int y=0;y<height;y++) {
			extractRow( imageReal , y ,  rowReal );	
			extractRow( imageImag , y ,  rowImag );	
			doFFT1D(rowReal, rowImag);
			putRow( imageReal , y ,  rowReal );	
			putRow( imageImag , y ,  rowImag );	

		}
	}

	/*------------------------------------------------------------------*/
	/**
	 * Perform an inverse FFT 2D.
	 * Each image's dimension has to be a power of two : 2**m
	 *
	 * @param imageReal  input: Fourier coefficients (real part) , output: Signal (real part) 
	 * @param imageImag  input: Fourier coefficients (imaginary part), output: Signal (imaginary part)
	 * @param width		width of the signal
	 * @param height		height of the signal
	 */
	static public void inverseFFT( double[] imageReal, double[] imageImag, int width, int height) {
		double colReal[] = new double[height];
		double colImag[] = new double[height];
		for (int x=0;x<width;x++) {
			extractColumn( imageReal , width , x , colReal );	
			extractColumn( imageImag , width , x , colImag );						
			inverseFFT1D( colReal , colImag );
			putColumn( imageReal , width , x , colReal );	
			putColumn( imageImag , width , x , colImag );
		}

		double rowReal[] = new double[width];
		double rowImag[] = new double[width];
		for (int y=0;y<height;y++) {
			extractRow( imageReal , y ,  rowReal );	
			extractRow( imageImag , y ,  rowImag );	
			inverseFFT1D(rowReal, rowImag);
			putRow( imageReal , y ,  rowReal );	
			putRow( imageImag , y ,  rowImag );	

		}
	}

	/*....................................................................
	private methods
....................................................................*/

	/*------------------------------------------------------------------*/
	static private void extractColumn ( final double[] array, final int width, int x, final double[] column ) {
		for (int k = 0, K = column.length; (k < K); k++) {
			column[k] = (double)array[x];
			x += width;
		}
	} /* end extractColumn */

	/*------------------------------------------------------------------*/
	static private void extractRow ( final double[] array, int y, final double[] row ) {
		y *= row.length;
		for (int k = 0, K = row.length; (k < K); k++) {
			row[k] = (double)array[y++];
		}
	} /* end extractRow */

	/*------------------------------------------------------------------*/
	static private void putColumn ( final double[] array, final int width, int x, final double[] column) {
		for (int k = 0, K = column.length; (k < K); k++) {
			array[x] = (float)column[k];
			x += width;
		}
	} /* end putColumn */

	/*------------------------------------------------------------------*/
	static private void putRow ( final double[] array, int y, final double[] row) {
		y *= row.length;
		for (int k = 0, K = row.length; (k < K); k++) {
			array[y++] = (float)row[k];
		}
	} /* end putRow */

	/*------------------------------------------------------------------*/
	/**
	 * Perform the FFT 1D Method Decimation in time (Cooley Tukey)
	 *
	 * Signal's size has to be a power of two : 2**m
	 */
	static private void doFFT1D(double real[], double imaginary[]) {
		int shift = 0; 				//start of the computation
		int size = real.length;		//length of the FFT

		int m = (int)Math.floor((Math.log((double)size)/Math.log((double)2)));

		double	Retmp, Imtmp, arg;
		int i,j,k,stepsize,shifter,n;
		int i_j,i_j_s;

		n=1<<m;

		double[]Imarg = new double[n];
		double[]Rearg = new double[n];

		//compute W coefficients
		double arg0 = 2.0 * Math.PI / (float)n;
		for (i=0;i<n;i++) {
			arg = arg0 * (float)i;
			Rearg[i] = Math.cos(arg);
			Imarg[i] = -Math.sin(arg);
		}

		//bit inversion	
		for (i=j=shift;i<shift+n-1;i++) {
			if (i<j) {
				Retmp=real[i]; Imtmp=imaginary[i];
				real[i]=real[j]; imaginary[i]=imaginary[j];
				real[j]=Retmp; imaginary[j]=Imtmp;
			}
			k=n>>1;
			while (k+shift<=j) {
				j-=k;
				k/=2;
			}
			j+=k;
		}

		//Perform the FFT
		for(stepsize=1,shifter=m-1; stepsize<n; stepsize<<=1,--shifter)
		{
			for(j=shift;j<shift+n;j+=stepsize<<1) {
				for (i=0;i<stepsize;i++) {
					i_j=i+j;
					i_j_s=i_j+stepsize;
					if(i>0) {
						Retmp = Rearg[i<<shifter]*real[i_j_s]-Imarg[i<<shifter]*imaginary[i_j_s];
						imaginary[i_j_s] = Rearg[i<<shifter]*imaginary[i_j_s]+Imarg[i<<shifter]*real[i_j_s];
						real[i_j_s] = Retmp;
					}
					Retmp = real[i_j] - real[i_j_s];
					Imtmp = imaginary[i_j] - imaginary[i_j_s];
					real[i_j] += real[i_j_s];
					imaginary[i_j] += imaginary[i_j_s];
					real[i_j_s] = Retmp;
					imaginary[i_j_s] = Imtmp;
				}
			}
		}
	}

	/*------------------------------------------------------------------*/
	/**
	 * Perform the Inverse of the FFT 1D.
	 * Signal's size has to be a power of two : 2**m
	 */
	static private void inverseFFT1D(double real[], double imaginary[]) {
		int shift = 0; 				//start of the computation
		int size = real.length; 	//length of the IFFT

		for (int i=shift;i<shift+size;i++) {
			imaginary[i] = -imaginary[i];
		}

		doFFT1D(real, imaginary);

		for (int i=shift;i<shift+size;i++) {
			real[i] = real[i]/size;
			imaginary[i] = -imaginary[i]/size;
		}
	}
} // End FFTTools